@@ -1,20 +1,16 @@
-# -*- coding: utf-8 -*-
-
 import logging
 import os
 import shutil
 import unittest
 import uuid
 from datetime import datetime, timedelta, timezone
-
-from pathlib import Path
-
 from io import BytesIO
+from pathlib import Path
 
 from pykeepass import PyKeePass, icons
 from pykeepass.entry import Entry
-from pykeepass.group import Group
 from pykeepass.exceptions import BinaryError, CredentialsError, HeaderChecksumError
+from pykeepass.group import Group
 
 """
 Missing Tests:
@@ -247,7 +243,7 @@ class EntryFindTests3(KDBX3Tests):
         self.assertEqual(results.url, unique_str + 'url')
         self.assertEqual(results.notes, unique_str + 'notes')
         self.assertEqual(len(results.tags), 6)
-        self.assertTrue(results.uuid != None)
+        self.assertTrue(results.uuid is not None)
         self.assertTrue(results.autotype_sequence is None)
         self.assertEqual(results.icon, icons.KEY)
 
@@ -402,7 +398,7 @@ class GroupFindTests3(KDBX3Tests):
         results = self.kp.find_groups(path=['base_group', 'sub_group'], first=True)
         self.assertIsInstance(results, Group)
         self.assertEqual(results.name, sub_group.name)
-        self.assertTrue(results.uuid != None)
+        self.assertTrue(results.uuid is not None)
 
         self.kp.move_group(sub_group2, sub_group)
         results = self.kp.find_groups(path=['base_group', 'sub_group', 'sub_group2'], first=True)
@@ -532,6 +528,20 @@ class EntryTests3(KDBX3Tests):
         self.assertNotEqual(original_entry, clone1)
         self.assertNotEqual(clone1, clone2)
 
+    def test_broken_reference(self):
+        # TODO: move the entry into test databases
+        broken_entry_title = 'broken reference'
+        self.kp.add_entry(
+            self.kp.root_group,
+            title=broken_entry_title,
+            username='{REF:U@I:AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA}',
+            password='{REF:P@I:AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA}',
+        )
+        broken_entry = self.kp.find_entries(title=broken_entry_title, first=True)
+        self.assertEqual(broken_entry.deref('username'), None)
+        self.assertEqual(broken_entry.deref('password'), None)
+        self.kp.delete_entry(broken_entry)
+
     def test_set_and_get_fields(self):
         time = datetime.now(timezone.utc).replace(microsecond=0)
         changed_time = time + timedelta(hours=9)
@@ -582,6 +592,8 @@ class EntryTests3(KDBX3Tests):
         self.assertEqual(entry.tags, ['changed', 'tags'])
         entry.tags = ['changed', 'again', 'tags']
         self.assertEqual(entry.tags, ['changed', 'again', 'tags'])
+        entry.tags = []
+        self.assertEqual(entry.tags, [])
 
     def test_expired_datetime_offset(self):
         """Test for https://github.com/pschmitt/pykeepass/issues/115"""
@@ -879,6 +891,20 @@ class AttachmentTests3(KDBX3Tests):
 
 
 class PyKeePassTests3(KDBX3Tests):
+    def test_consecutives_saves_with_stream(self):
+        # https://github.com/libkeepass/pykeepass/pull/388
+        self.setUp()
+
+        with open(base_dir / self.keyfile_tmp, 'rb') as f:
+            keyfile = BytesIO(f.read())
+
+        for _i in range(5):
+            with PyKeePass(
+                base_dir / self.database_tmp,
+                password=self.password,
+                keyfile=keyfile,
+            ) as kp:
+                kp.save()
 
     def test_set_credentials(self):
         self.kp_tmp.password = 'f00bar'
@@ -1023,6 +1049,38 @@ class BugRegressionTests3(KDBX3Tests):
         e._element.xpath('Times/ExpiryTime')[0].text = None
         self.assertEqual(e.expiry_time, None)
 
+    def test_issue376(self):
+        # Setting the properties of an entry should not change the Protected
+        # property
+        subgroup = self.kp.root_group
+        e = self.kp.add_entry(subgroup, 'banana_entry', 'user', 'pass')
+
+        self.assertEqual(e._is_property_protected('Password'), True)
+        self.assertEqual(e._is_property_protected('Title'), False)
+        self.assertEqual(e.otp, None)
+        self.assertEqual(e._is_property_protected('otp'), False)
+
+        e.title = 'pineapple'
+        e.password = 'pizza'
+        e.otp = 'aa'
+
+        self.assertEqual(e._is_property_protected('Password'), True)
+        self.assertEqual(e._is_property_protected('Title'), False)
+        self.assertEqual(e._is_property_protected('otp'), True)
+
+        # Using protected=None should not change the current status
+        e._set_string_field('XYZ', '1', protected=None)
+        self.assertEqual(e._is_property_protected('XYZ'), False)
+
+        e._set_string_field('XYZ', '1', protected=True)
+        self.assertEqual(e._is_property_protected('XYZ'), True)
+
+        e._set_string_field('XYZ', '1', protected=None)
+        self.assertEqual(e._is_property_protected('XYZ'), True)
+
+        e._set_string_field('XYZ', '1', protected=False)
+        self.assertEqual(e._is_property_protected('XYZ'), False)
+
 class EntryFindTests4(KDBX4Tests, EntryFindTests3):
     pass
 
@@ -1054,8 +1112,39 @@ class CtxManagerTests(unittest.TestCase):
             results = kp.find_entries_by_username('foobar_user', first=True)
             self.assertEqual('foobar_user', results.username)
 
+class PyKeePassTests3(KDBX3Tests):
+    """Tests on PyKeePass class that don't involve attachments or finding entries/groups"""
+
+    def test_database_info(self):
+        """Test database properties"""
+
+        # Test name
+        self.assertEqual(self.kp_tmp.database_name, None)
+        self.kp_tmp.database_name = "Test Name"
+        self.assertEqual(self.kp_tmp.database_name, "Test Name")
+
+        # Test Description
+        self.assertEqual(self.kp_tmp.database_description, None)
+        self.kp_tmp.database_description = "Test Description"
+        self.assertEqual(self.kp_tmp.database_description, "Test Description")
+
+        # Test Default User Name
+        self.assertEqual(self.kp_tmp.default_username, None)
+        self.kp_tmp.default_username = "Test User"
+        self.assertEqual(self.kp_tmp.default_username, "Test User")
+
+        self.kp_tmp.save()
+        self.kp_tmp.reload()
+
+        self.assertEqual(self.kp_tmp.database_name, "Test Name")
+        self.assertEqual(self.kp_tmp.database_description, "Test Description")
+        self.assertEqual(self.kp_tmp.default_username, "Test User")
+
+class PyKeePassTests4(KDBX4Tests, PyKeePassTests3):
+    pass
 
 class KDBXTests(unittest.TestCase):
+    """Tests on KDBX parsing"""
 
     def test_open_save(self):
         """try to open all databases, save them, then open the result"""
